@@ -62,7 +62,7 @@
 static char version_string[] = "Modern traceroute for Linux, "
 				"version " _TEXT(VERSION) ", " __DATE__
 				"\nCopyright (c) 2006  Dmitry Butskoy, "
-				"  License: GPL";
+				"  License: GPL v2 or any later";
 static int debug = 0;
 static unsigned int first_hop = 1;
 static unsigned int max_hops = DEF_HOPS;
@@ -581,7 +581,7 @@ int main (int argc, char *argv[]) {
 		    exit (2);
 	}
 
-	if (ops->init (dst_name, &dst_addr, dst_port_seq, packet_len) < 0)
+	if (ops->init (&dst_addr, dst_port_seq, packet_len) < 0)
 		ex_error ("trace method's init failed");
 
 
@@ -804,9 +804,36 @@ replace_by_final:
 }
 
 
+probe *probe_by_seq (int seq) {
+	int n;
+
+	if (seq <= 0)  return NULL;
+
+	for (n = 0; n < num_probes; n++) {
+	    if (probes[n].seq == seq)
+		    return &probes[n];
+	}
+
+	return NULL;
+}
+
+probe *probe_by_sk (int sk) {
+	int n;
+
+	if (sk <= 0)  return NULL;
+
+	for (n = 0; n < num_probes; n++) {
+	    if (probes[n].sk == sk)
+		    return &probes[n];
+	}
+
+	return NULL;
+}
+
+
 static void poll_callback (int fd, int revents) {
 
-	ops->recv_probe (fd, revents, probes, num_probes);
+	ops->recv_probe (fd, revents);
 }
 
 
@@ -862,6 +889,11 @@ static void do_it (void) {
 
 		    ops->send_probe (pb, ttl);
 
+		    if (!pb->send_time) {
+			if (max_time)  break;	/*  have chances later   */
+			else  error ("send probe");
+		    }
+
 		    last_send = pb->send_time;
 		}
 
@@ -1245,11 +1277,29 @@ void set_ttl (int sk, int ttl) {
 }
 
 
-/*  There is a typo in the kernel before 2.6.22.2, which prevents icmp errors
+int do_send (int sk, const void *data, size_t len, const sockaddr_any *addr) {
+	int res;
+
+	if (!addr || raw_can_connect ())
+		res = send (sk, data, len, 0);
+	else
+	    res = sendto (sk, data, len, 0, &addr->sa, sizeof (*addr));
+
+	if (res < 0 &&
+	    errno != ENOBUFS && errno != EAGAIN
+	)  error ("send");	/*  not recoverable   */
+
+	return res;
+}
+
+
+/*  There is a bug in the kernel before 2.6.25, which prevents icmp errors
   to be obtained by MSG_ERRQUEUE for ipv6 connected raw sockets.
 */
 static int can_connect = -1;
 
+#define VER(A,B,C,D)	(((((((A) << 8) | (B)) << 8) | (C)) << 8) | (D))
+
 int raw_can_connect (void) {
 
 	if (can_connect < 0) {
@@ -1258,11 +1308,14 @@ int raw_can_connect (void) {
 		    can_connect = 1;
 	    else {	/*  AF_INET6   */
 		struct utsname uts;
+		int n;
+		unsigned int a, b, c, d = 0;
 
 		if (uname (&uts) < 0)
 			return 0;
 
-		can_connect = (strcmp (uts.release, "2.6.22.2") >= 0);
+		n = sscanf (uts.release, "%u.%u.%u.%u", &a, &b, &c, &d);
+		can_connect = (n >= 3 && VER (a, b, c, d) >= VER (2, 6, 25, 0));
 	    }
 	}
 
